import subprocess
import os
import numpy as np
import yaml
import uproot
from sigfig import round
__author__ = "Mario Krüger"
# -------------------------------------------------------------------------------------------------------------------------------------


def getDataPoints(file, histName):
    """
    extract all datapoints of a histogram in form of values, statErrors, systErrors, binEdges
    """
    statHist = file[histName]
    systHist = file[histName + "_syst"]
    print("- extracting histogram {} of type {}".format(statHist.member("fName"), statHist.classname))
    histArrays = statHist.to_numpy()
    values = histArrays[0]
    statErrors = statHist.errors()
    systErrors = systHist.errors()
    binEdges = [histArrays[1]]
    if (values.ndim == 2):
        binEdges.append(histArrays[2])

    return [values, statErrors, systErrors, binEdges]
# -------------------------------------------------------------------------------------------------------------------------------------


def rootLateXToProperLateX(text):
    """
    undo the ROOT LateX peculiarities
    """
    text = "$" + text.replace('#', '\\') + "$"
    text = text.replace(r'LT', r'langle')
    text = text.replace(r'GT', r'rangle')
    text = text.replace(r'ch', r'\mathrm{ch}')
    text = text.replace(r' (GeV', r'\ (\mathrm{GeV}')
    text = text.replace(r'\it{N}', r'N')
    text = text.replace(r'\it{P}', r'P')
    text = text.replace(r'\it{p}', r'p')
    text = text.replace(r'\it{c}', r'c')
    text = text.replace(r'\it{\eta}', r'\eta')
    text = text.replace(r'T', r'\mathrm{T}')
    text = text.replace(r'd', r'\mathrm{d}')
    text = text.replace(r'evt', r'\mathrm{evt}')
    text = text.replace(r'incl', r'\mathrm{incl}')
    text = text.replace(r'1/N_{\mathrm{evt}, N_{\mathrm{ch}}> 0} ', r'1/N_{\mathrm{evt}, N_{\mathrm{ch}}> 0}\ ')
    return text
# -------------------------------------------------------------------------------------------------------------------------------------


def histToHEP(file, histName):
    """
    convert root histogram to hep data entry
    """
    roundNumbers = True
    values, statErrors, systErrors, binEdges = getDataPoints(file, histName)
    hist = file[histName]

    # extract independent variables
    independent_variables = []
    for dim in range(0, len(binEdges)):
        if (dim == 0):
            axis = hist.member("fXaxis")
        else:
            axis = hist.member("fYaxis")
        ind_name = axis.member("fTitle")
        ind_name = rootLateXToProperLateX(ind_name)
        independent_variables.append({"header": {"name": ind_name}, "values": []})

        nLoopRepetitions = 1
        nLineRepetitions = 1

        if (len(binEdges) == 2):
            if (dim == 0):
                nLineRepetitions = binEdges[1].size-1
            if (dim == 1):
                nLoopRepetitions = binEdges[0].size-1

        for k in range(0, nLoopRepetitions):
            for i in range(1, binEdges[dim].size):
                low = "{}".format(binEdges[dim][i-1])
                high = "{}".format(binEdges[dim][i])
                for j in range(0, nLineRepetitions):
                    if roundNumbers:
                        low = round(low, sigfigs=5, output_type=float)
                        high = round(high, sigfigs=5, output_type=float)
                    independent_variables[dim]["values"].append({"low": float(low), "high": float(high)})

    # extract dependent variables
    dependent_variables = []
    dep_name = hist.member("fYaxis").member("fTitle")
    if (values.ndim == 2):
        dep_name = hist.member("fZaxis").member("fTitle")
    dep_name = rootLateXToProperLateX(dep_name)
    dependent_variables.append({"header": {"name": dep_name}, "values": []})
    for value, statError, systError in zip(values, statErrors, systErrors):
        if (values.ndim == 1):
            value = [value]
            statError = [statError]
            systError = [systError]
        for val, statErr, systErr in zip(value, statError, systError):
            if (val == 0 and statErr == 0 and systErr == 0):
                dependent_variables[0]["values"].append({"value": "-", "errors": []})
            else:
                if roundNumbers:
                    smallErr = statErr
                    #print(val, systErr, statErr)
                    val, statErr = round(val, smallErr, cutoff=29, sep=tuple)
                    systErr, _ = round(systErr, smallErr, cutoff=29, sep=tuple)
                    #print(val, systErr, statErr)
                dependent_variables[0]["values"].append({"value": float(val), "errors": [{"symerror": float(statErr), "label": "stat"}, {"symerror": float(systErr), "label": "syst"}]})

    # remove empty bins of the histogram in both dependent and independent variables
    emptyBins = [i for i in range(len(dependent_variables[0]["values"])) if dependent_variables[0]["values"][i]["value"] == "-"]
    for i in sorted(emptyBins, reverse=True):
        del independent_variables[0]["values"][i]
        if (values.ndim == 2):
            del independent_variables[1]["values"][i]
        del dependent_variables[0]["values"][i]

    hepEntry = {"independent_variables": independent_variables,
                "dependent_variables": dependent_variables}

    return hepEntry
# -------------------------------------------------------------------------------------------------------------------------------------


def makeHEPDataEntry():
    """
    make the hep data entry for our measurement
    """
    outDir = "MultDepSpecHEPData/"
    if not os.path.exists(outDir):
        os.makedirs(outDir)

    file = uproot.open("SpectralShapeEvolution.root")

    collisionSystems = ("pp", "pPb", "XeXe", "PbPb")
    collisionSystemsLabels = ("P P", "P PB", "XE XE", "PB PB")
    energies = [["2", "5", "7", "8", "13"], ["5", "8"], ["5"], ["2", "5"]]
    energiesGeV = [[2760, 5020, 7000, 8000, 13000], [5020, 8160], [5440], [2760, 5020]]
    dataNames = ("multDist", "multDistKNO", "multPtSpec", "multMeanPt", "multStdDevPt", "multRelStdDevPt", "ptSpec", "multPtSpec_scaled", "multMeanPt_scaled", "multStdDevPt_scaled", "multRelStdDevPt_scaled", "multMeanPtOverIncl_scaled", "multStdDevPtOverIncl_scaled")
    dataNamesLabels = ("MULT", "MULTKNO", "D3N/DETARAP/DPT/DMULT", "MEANPT", "STDDEVPT", "STDDEVPTOVERMEANPT", "D2N/DETARAP/DPT", "D3N/DETARAP/DPT/DMULT", "MEANPT", "STDDEVPT", "STDDEVPTOVERMEANPT", "MEANPT", "STDDEVPT")
    dataDescriptions = ("Charged-particle multiplicity distribution",
                        "Koba-Nielsen-Olesen scaled charged-particle multiplicity distribution",
                        "Charged-particle transverse momentum spectra as a function of charged-particle multiplicity",
                        "Charged-particle mean transverse momentum as a function of charged-particle multiplicity",
                        "Standard deviation of charged-particle transverse momentum spectra as a function of charged-particle multiplicity",
                        "Standard deviation over mean of charged-particle transverse momentum spectra as a function of charged-particle multiplicity",
                        "Charged-particle transverse momentum spectra integrated over charged-particle multipliciy",
                        "Charged-particle transverse momentum spectra as a function of relative charged-particle multiplicity",
                        "Charged-particle mean transverse momentum as a function of relative charged-particle multiplicity",
                        "Standard deviation of charged-particle transverse momentum spectra as a function of relative charged-particle multiplicity",
                        "Standard deviation over mean of charged-particle transverse momentum spectra as a function of relative charged-particle multiplicity",
                        "Charged-particle mean transverse momentum divided by its multiplicity-integrated value as a function of relative charged-particle multiplicity",
                        "Standard deviation of charged-particle transverse momentum spectra divided by its multiplicity-integrated value as a function of relative charged-particle multiplicity"
                        )

    figures = []
    listOfTables = []
    colSysID = 0
    for colSys, colSysLabel in zip(collisionSystems, collisionSystemsLabels):
        for energy, energyGeV in zip(energies[colSysID], energiesGeV[colSysID]):
            folderName = colSys + "_" + energy + "TeV"
            print(folderName + ":")
            for dataName, dataNameLabel, dataDescription in zip(dataNames, dataNamesLabels, dataDescriptions):

                keywords = [
                    {"name": "reactions", "values": [colSysLabel + " --> CHARGED X"]},
                    {"name": "observables", "values": [dataNameLabel]},
                    {"name": "cmenergies", "values": [energyGeV]}]
                # additional_resources = [{"description": "Figure from paper.", "location": dataName + "_" + (colSys if colSys != "XeXe" and colSys != "PbPb" else "AA") + ".png"} if "tSpec" not in dataName and "scaled" not in dataName and "Rel" not in dataName else {}]
                table_info = {"name": folderName + "_" + dataName, "description": dataDescription + " for {} collisions at {} TeV.".format(colSys, energyGeV/1000),
                              "keywords": keywords, "data_file": "data_" + folderName + "_{}.yaml".format(dataName),
                              # "additional_resources": additional_resources
                              }
                tablePath = outDir + "table_" + folderName + "_{}.yaml".format(dataName)
                listOfTables.append(tablePath)

                # if additional_resources[0] != {}:
                #    figures.append(additional_resources[0]["location"])
                # else:
                #    del table_info["additional_resources"]

                with open(tablePath, "w") as data_stream:
                    yaml.dump(table_info, data_stream, sort_keys=False, width=float("inf"))

                hepEntry = histToHEP(file, folderName + "/" + dataName)
                hepEntry["dependent_variables"][0]["qualifiers"] = [
                    {"name": "RE", "value": colSysLabel + " --> CHARGED X"},
                    {"name": "SQRT(S)" + ("/NUCLEON" if colSys != "pp" else ""), "units": "GEV", "value": energyGeV},
                    {"name": "ETARAP", "value": "-0.8 TO 0.8"}, {"name": "PT", "units": "GEV/C", "value": "0.15 TO 10"}]

                with open(outDir + "data_" + folderName + "_{}.yaml".format(dataName), "w") as data_stream:
                    yaml.dump(hepEntry, data_stream, sort_keys=False, width=float("inf"))

        colSysID += 1

    # generate the submission file
    meta_infos = {"additional_resources": [{"location": "SpectralShapeEvolution.root", "description": "analysis output file containing all data shown here"}, {"location": "makeHEPDataEntry.py", "description": "python script used to create this HEPData entry from the attached root file"}],
                  "comment":
                  r"CERN-LHC. ALICE. "
                  r"Multiplicity ($N_{\mathrm{ch}}$) distributions and transverse momentum ($p_{\mathrm{T}}$) spectra of inclusive primary charged particles in the kinematic range of $|\eta| < 0.8$ and $0.15\ \mathrm{GeV}/c < p_{\mathrm{T}} < 10\ \mathrm{GeV}/c$ are reported for pp, p-Pb, Xe-Xe and Pb-Pb collisions at centre-of-mass energies per nucleon pair ranging from $\sqrt{s_\mathrm{NN}} = 2.76\ \mathrm{TeV}$ up to $13\ \mathrm{TeV}$. "
                  r"The multiplicity distributions are also shown in their KNO-scaled form. "
                  r"The correlation between the transverse momentum of primary charged particles and the charged-particle multiplicity of the corresponding collision is reported in fine intervals. "
                  r"In addition, the mean $ \langle p_{\mathrm{T}} \rangle $ and standard deviation $ \sigma(p_{\mathrm{T}}) $ derived from these multiplicity dependent $p_\mathrm{T}$ spectra are shown as a function of $N_{\mathrm{ch}}$. "
                  r"The relative standard deviation $ \sigma(p_{\mathrm{T}}) / \langle p_{\mathrm{T}} \rangle $ of the spectra is reported as well. "
                  r"All observables are also presented as a function of the reltaive multiplicity $ N_{\mathrm{ch}}\ / \langle N_{\mathrm{ch}} \rangle $. "
                  r"Note that $ \langle N_{\mathrm{ch}} \rangle $ is calculated for the whole multiplicity range extending beyond the $N_{\mathrm{ch}}$ values reported for the multiplicity-differential observables. "
                  r"Also the transverse momentum spectra integrated over this full range of multiplicities are presented. "
                  r"The mean and standard deviation derived from these multiplicity-integrated $p_\mathrm{T}$ spectra are used to normalize the corresponding multiplicity-dependent values which are then reported as a function of $ N_{\mathrm{ch}}\ / \langle N_{\mathrm{ch}} \rangle $."
                  }

    with open(outDir + "submission.yaml", "w") as data_stream:
        yaml.dump(meta_infos, data_stream, sort_keys=False, width=float("inf"))

    myString = "---\n"
    bashCommand = "cat " + outDir + "submission.yaml"
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()
    myString += output.decode("utf-8")

    for tablePath in listOfTables:
        # concat the submission file and the individual table files into one and delete the individual files
        bashCommand = "cat " + tablePath
        process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()
        myString += "---\n" + output.decode("utf-8")
        bashCommand = "rm " + tablePath
        process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()

    submissionFile = open(outDir + "submission.yaml", 'w')
    submissionFile.write(myString)
    submissionFile.close()

    # copy root file and this script into the outuput directory
    bashCommand = "cp SpectralShapeEvolution.root makeHEPDataEntry.py " + outDir
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()

    # for figure in figures:
    #    bashCommand = "cp previews/" + figure + " " + outDir
    #    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    #    output, error = process.communicate()
# -------------------------------------------------------------------------------------------------------------------------------------
if __name__ == "__main__":
    makeHEPDataEntry()
